@@ -40,11 +40,87 @@
 #include <scsi/sg.h>
 #endif /* __linux__ */
 
+#ifdef __FreeBSD__
+#include <sys/cdio.h>
+#include <sys/cdrio.h>
+#include <camlib.h>
+#endif /* __FreeBSD__ */
+
 #include <glib.h>
 #include <libgnome/gnome-i18n.h>
 
 #include "cd-drive.h"
 
+#if defined(__linux__) || defined(__FreeBSD__)
+
+/* For dvd_plus_rw_utils.cpp */
+int get_dvd_r_rw_profile (const char *name);
+int get_mmc_profile (void *fd);
+
+static void
+add_dvd_plus (CDDrive *cdrom)
+{
+	int caps;
+
+	caps = get_dvd_r_rw_profile (cdrom->device);
+
+	if (caps == -1) {
+		return;
+	}
+
+	if (caps == 2) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_RW_RECORDER;
+		cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_R_RECORDER;
+	} else if (caps == 0) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_R_RECORDER;
+	} else if (caps == 1) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_RW_RECORDER;
+	}
+}
+
+static CDMediaType
+linux_bsd_media_type (const char *device)
+{
+	int fd;
+	int mmc_profile;
+
+	fd = open (device, O_RDONLY);
+	if (fd < 0) {
+		return CD_MEDIA_TYPE_UNKNOWN;
+	}
+
+	mmc_profile = get_mmc_profile ((void *)fd);
+	close (fd);
+
+	switch (mmc_profile) {
+	case -1:        /* Couldn't get the data about the media */
+		return CD_MEDIA_TYPE_UNKNOWN;
+	case 0x8:	/* Commercial CDs and Audio CD	*/
+		return CD_MEDIA_TYPE_CD;
+	case 0x9:	/* CD-R                         */
+		return CD_MEDIA_TYPE_CDR;
+	case 0xa:	/* CD-RW			*/
+		return CD_MEDIA_TYPE_CDRW;
+	case 0x10:	/* Commercial DVDs		*/
+		return CD_MEDIA_TYPE_DVD;
+	case 0x11:      /* DVD-R                        */
+		return CD_MEDIA_TYPE_DVDR;
+	case 0x13:      /* DVD-RW Restricted Overwrite  */
+	case 0x14:      /* DVD-RW Sequential            */
+		return CD_MEDIA_TYPE_DVDRW;
+	case 0x1B:      /* DVD+R                        */
+		return CD_MEDIA_TYPE_DVD_PLUS_R;
+	case 0x1A:      /* DVD+RW                       */
+		return CD_MEDIA_TYPE_DVD_PLUS_RW;
+	case 0x12:      /* DVD-RAM                      */
+		return CD_MEDIA_TYPE_DVD_RAM;
+	default:
+		return CD_MEDIA_TYPE_UNKNOWN;
+	}
+}
+
+#endif /* __linux__ || __FreeBSD__ */
+
 #ifdef __linux__
 
 #ifdef USE_STABLE_LIBGLIB
@@ -100,7 +176,6 @@ read_lines (char *filename)
 }
 
 struct scsi_unit {
-	gboolean exist;
 	char *vendor;
 	char *model;
 	char *rev;
@@ -125,23 +200,22 @@ get_scsi_units (char **device_str, char **devices, struct scsi_unit *scsi_units)
 {
 	char vendor[9], model[17], rev[5];
 	int host_no, access_count, queue_depth, device_busy, online, channel;
+	int scsi_id, scsi_lun, scsi_type;
 	int i, j;
 
 	for (i = 0, j = 0; device_str[i] != NULL && devices[i] != NULL; i++) {
 		if (strcmp (device_str[i], "<no active device>") == 0) {
-			scsi_units[i].exist = FALSE;
 			continue;
 		}
 		if (sscanf (devices[i], "%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d",
 			    &host_no,
-			    &channel, &scsi_units[j].id, &scsi_units[j].lun, 
-				&scsi_units[j].type, &access_count, &queue_depth,
+			    &channel, &scsi_id, &scsi_lun, &scsi_type, &access_count, &queue_depth,
 				&device_busy, &online) != 9) {
 		
 			g_warning ("Couldn't match line in /proc/scsi/sg/devices\n");
 			continue;
 		}
-		if (scsi_units[j].type == 5) { /* TYPE_ROM (include/scsi/scsi.h) */
+		if (scsi_type == 5) { /* TYPE_ROM (include/scsi/scsi.h) */
 			if (sscanf (device_str[i], "%8c\t%16c\t%4c", 
 						vendor, model, rev) != 3) {
 				g_warning ("Couldn't match line /proc/scsi/sg/device_strs\n");
@@ -153,7 +227,10 @@ get_scsi_units (char **device_str, char **devices, struct scsi_unit *scsi_units)
 			scsi_units[j].model = g_strdup (g_strstrip (model));
 			scsi_units[j].rev = g_strdup (g_strstrip (rev));
 			scsi_units[j].bus = host_no;
-			scsi_units[j].exist = TRUE;
+			scsi_units[j].id = scsi_id;
+			scsi_units[j].lun = scsi_lun; 
+			scsi_units[j].type = scsi_type;
+			
 			j++;
 		}
 	}
@@ -259,12 +336,18 @@ get_device_max_speed (char *id)
 		if (speed != NULL) {
 			speed += strlen ("Maximum write speed in kB/s:");
 			max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
+		} else {
+			speed = strstr (stdout_data, "Maximum write speed:");
+			if (speed != NULL) {
+				speed += strlen ("Maximum write speed:");
+				max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
+			}
 		}
+		g_free (stdout_data);
 	}
 
 	g_free (dev_str);
 	return max_speed;
-
 }
 
 
@@ -276,7 +359,7 @@ get_scsi_cd_name (int bus, int id, int lun, const char *dev,
 	
 	scsi_unit = lookup_scsi_unit (bus, id, lun, scsi_units, n_scsi_units);
 	if (scsi_unit == NULL) {
-		return g_strdup_printf (_("Unnamed SCSI CDROM (%s)"), dev);
+		return g_strdup_printf (_("Unnamed SCSI CD-ROM (%s)"), dev);
 	}
 
 	return g_strdup_printf ("%s - %s",
@@ -332,6 +415,7 @@ cdrom_get_name (struct cdrom_unit *cdrom, struct scsi_unit *scsi_units, int n_sc
 
 static GList *
 add_linux_cd_recorder (GList *cdroms,
+		       gboolean recorder_only,
 		       struct cdrom_unit *cdrom_s,
 		       struct scsi_unit *scsi_units,
 		       int n_scsi_units)
@@ -341,6 +425,8 @@ add_linux_cd_recorder (GList *cdroms,
 
 	cdrom = g_new0 (CDDrive, 1);
 
+	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
+
 	if (cdrom_s->device[0] == 's') {
 		cdrom->display_name = cdrom_get_name (cdrom_s, scsi_units,
 				n_scsi_units);
@@ -349,22 +435,44 @@ add_linux_cd_recorder (GList *cdroms,
 		}
 		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d",
 				bus, id, lun);
-		cdrom->max_speed_write = get_device_max_speed (cdrom->cdrecord_id);
 	} else {
 		/* kernel >=2.5 can write cd w/o ide-scsi */
-		cdrom->display_name = cdrom_get_name(cdrom_s, scsi_units, n_scsi_units);
-		cdrom->cdrecord_id = g_strdup_printf (cdrom_s->device);
-		cdrom->max_speed_write = get_device_max_speed (cdrom->cdrecord_id);
+		cdrom->display_name = cdrom_get_name (cdrom_s,
+				scsi_units, n_scsi_units);
+		cdrom->cdrecord_id = g_strdup_printf ("/dev/%s",
+				cdrom_s->device);
 	}
-	cdrom->device = g_strdup_printf ("/dev/%s", cdrom_s->device);
-	cdrom->max_speed_read = cdrom_s->speed; 
-	if (cdrom_s->can_write_dvdr
-	    || cdrom_s->can_write_dvdram) {
-		cdrom->type = CDDRIVE_TYPE_DVD_RECORDER;
+
+	if (recorder_only) {
+		cdrom->max_speed_write = get_device_max_speed
+			(cdrom->cdrecord_id);
 	} else {
-		cdrom->type = CDDRIVE_TYPE_CD_RECORDER;
+		/* Have a wild guess, the drive should actually correct us */
+		cdrom->max_speed_write = cdrom_s->speed;
 	}
-	
+
+	cdrom->device = g_strdup_printf ("/dev/%s", cdrom_s->device);
+	cdrom->max_speed_read = cdrom_s->speed;
+	if (cdrom_s->can_write_dvdr) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
+	}
+
+	if (cdrom_s->can_write_dvdram) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
+	}
+
+	if (cdrom_s->can_write_cdr) {
+		cdrom->type |= CDDRIVE_TYPE_CD_RECORDER;
+	}
+	if (cdrom_s->can_write_cdrw) {
+		cdrom->type |= CDDRIVE_TYPE_CDRW_RECORDER;
+	}
+	if (cdrom_s->can_read_dvd) {
+		cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
+	}
+
+	add_dvd_plus (cdrom);
+
 	return g_list_append (cdroms, cdrom);
 }
 
@@ -375,17 +483,16 @@ add_linux_cd_drive (GList *cdroms, struct cdrom_unit *cdrom_s,
 	CDDrive *cdrom;
 
 	cdrom = g_new0 (CDDrive, 1);
+	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
 	cdrom->cdrecord_id = NULL;
 	cdrom->display_name = cdrom_get_name (cdrom_s, scsi_units, n_scsi_units);
 	cdrom->device = g_strdup_printf ("/dev/%s", cdrom_s->device);
 	cdrom->max_speed_write = 0; /* Can't write */
 	cdrom->max_speed_read = cdrom_s->speed;
 	if (cdrom_s->can_read_dvd) {
-		cdrom->type = CDDRIVE_TYPE_DVD_DRIVE;
-	} else {
-		cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
+		cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
 	}
-	
+
 	return g_list_append (cdroms, cdrom);
 }
 
@@ -414,16 +521,18 @@ linux_scan (gboolean recorder_only)
 	struct cdrom_unit *cdroms;
 	char *p, *t;
 	int n_cdroms, maj, min, i, j;
-	int n_scsi_units = 0;
+	int n_scsi_units;
 	int fd;
+	FILE *file;
 	GList *cdroms_list;
-	gboolean have_devfs = FALSE;
+	gboolean have_devfs;
 
 	/* devfs creates and populates the /dev/cdroms directory when its mounted
 	 * the 'old style names' are matched with devfs names below.
 	 * The cdroms.device string gets cleaned up again in cdrom_get_name()
 	 * we need the oldstyle name to get device->display_name for ide.
 	 */
+	have_devfs = FALSE;
 	if (g_file_test ("/dev/cdroms", G_FILE_TEST_IS_DIR)) {
 		have_devfs = TRUE;
 	}
@@ -458,6 +567,7 @@ linux_scan (gboolean recorder_only)
 	 * devices are listed in /proc/sys/dev/cdrom/info. It will always
 	 * be 'h' or 's'
 	 */
+	n_scsi_units = 0;
 	for (i = 0; i < n_cdroms; i++) {
 		if (cdroms[i].device[0] == 's') {
 			n_scsi_units++;
@@ -565,12 +675,12 @@ linux_scan (gboolean recorder_only)
 	g_strfreev (cdrom_info);
 
 	/* get kernel major.minor version */
-	(FILE *) fd = fopen("/proc/sys/kernel/osrelease", "r");
-	if ((FILE *) fd == NULL || fscanf((FILE *) fd, "%d.%d", &maj, &min) != 2) {
+	file = fopen("/proc/sys/kernel/osrelease", "r");
+	if (file == NULL || fscanf(file, "%d.%d", &maj, &min) != 2) {
 		g_warning("Could not get kernel version.");
 		maj = min = 0;
 	}
-	fclose((FILE *) fd);
+	fclose(file);
 	
 	cdroms_list = NULL;
 	for (i = n_cdroms - 1, j = 0; i >= 0; i--, j++) {
@@ -585,7 +695,8 @@ linux_scan (gboolean recorder_only)
 						cdroms[i].device,  j);
 			}
 			cdroms_list = add_linux_cd_recorder (cdroms_list,
-					&cdroms[i], scsi_units, n_scsi_units);
+					recorder_only, &cdroms[i],
+					scsi_units, n_scsi_units);
 		} else if (!recorder_only) {
 			if (have_devfs) {
 				cdroms[i].device = g_strdup_printf("%s cdroms/cdrom%d",
@@ -603,6 +714,87 @@ linux_scan (gboolean recorder_only)
 }
 #endif /* __linux__ */
 
+#ifdef __FreeBSD__
+static GList *
+freebsd_scan (gboolean recorder_only)
+{
+	GList *cdroms_list = NULL;
+	const char *dev_type = "cd";
+	int fd;
+	int speed = 16; /* XXX Hardcode the write speed for now. */
+	int max_speed = CDR_MAX_SPEED;
+	int i = 0;
+
+	while (1) {
+		CDDriveType type;
+		CDDrive *cdrom;
+		gchar *cam_path, *dev;
+		struct cam_device *cam_dev;
+
+		type = 0;
+		cam_path = g_strdup_printf ("/dev/%s%dc", dev_type, i);
+
+		if (!g_file_test (cam_path, G_FILE_TEST_EXISTS)) {
+			g_free (cam_path);
+			break;
+		}
+
+		if ((cam_dev = cam_open_spec_device (dev_type, i, O_RDWR, NULL)) == NULL) {
+			i++;
+			g_free (cam_path);
+			continue;
+		}
+
+		/* XXX Other controllers might need to be added. */
+		if ((strncmp (cam_dev->sim_name, "ata", 3)) == 0) {
+			dev = g_strdup_printf ("/dev/a%s%dc", dev_type, i);
+		} else {
+			dev = g_strdup (cam_path);
+		}
+
+		g_free (cam_path);
+
+		if ((fd = open (dev, O_RDWR, 0)) < 0) {
+			g_free (dev);
+			free (cam_dev);
+			i++;
+			continue;
+		}
+
+		type |= CDDRIVE_TYPE_CD_DRIVE;
+		if (ioctl (fd, CDRIOCWRITESPEED, &max_speed) >= 0) {
+			/* XXX is it CD-R or CD-RW ? */
+			type |= CDDRIVE_TYPE_CD_RECORDER;
+		}
+
+		close (fd);
+
+		if (type & CDDRIVE_TYPE_CD_RECORDER
+				|| type & CDDRIVE_TYPE_CDRW_RECORDER
+				|| recorder_only == FALSE) {
+			cdrom = g_new0 (CDDrive, 1);
+			cdrom->display_name = g_strdup_printf ("%s %s", cam_dev->inq_data.vendor, cam_dev->inq_data.revision);
+			cdrom->device = g_strdup (dev);
+			cdrom->max_speed_read = speed;
+			cdrom->max_speed_write = speed;
+			cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", cam_dev->path_id, cam_dev->target_id, cam_dev->target_lun);
+			cdrom->type = type;
+
+			add_dvd_plus (cdrom);
+
+			cdroms_list = g_list_append (cdroms_list, cdrom);
+		}
+
+		g_free (dev);
+		free (cam_dev);
+
+		i++;
+	}
+
+	return cdroms_list;
+}
+#endif /* __FreeBSD__ */
+
 GList *
 scan_for_cdroms (gboolean recorder_only, gboolean add_image)
 {
@@ -613,6 +805,10 @@ scan_for_cdroms (gboolean recorder_only, gboolean add_image)
 	cdroms = linux_scan (recorder_only);
 #endif
 
+#ifdef __FreeBSD__
+	cdroms = freebsd_scan (recorder_only);
+#endif
+
 	if (add_image) {
 		/* File */
 		cdrom = g_new0 (CDDrive, 1);
@@ -637,3 +833,13 @@ cd_drive_free (CDDrive *drive)
 	g_free (drive->device);
 }
 
+CDMediaType
+guess_media_type (const char *device_path)
+{
+#if defined(__linux__) || defined(__FreeBSD__)
+	return linux_bsd_media_type (device_path);
+#else
+	return CD_BLANK_MEDIA_TYPE_UNKNOWN;
+#endif
+}
+
