@@ -303,6 +303,87 @@ insert_rwcd_request_cb (CDRecorder *cdrecorder, gboolean is_reload,
 	}
 }
 
+static gboolean
+media_type_matches (CDMediaType type, gboolean blank_cd)
+{
+	switch (type)
+	{
+	case CD_MEDIA_TYPE_UNKNOWN:
+	case CD_MEDIA_TYPE_CD:
+	case CD_MEDIA_TYPE_DVD:
+	case CD_MEDIA_TYPE_DVD_RAM:
+		return FALSE;
+	case CD_MEDIA_TYPE_CDR:
+		return (blank_cd ? FALSE : TRUE);
+	case CD_MEDIA_TYPE_CDRW:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVDR:
+		return (blank_cd ? FALSE : TRUE);
+	case CD_MEDIA_TYPE_DVDRW:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVD_PLUS_R:
+		return (blank_cd ? FALSE : TRUE);
+	case CD_MEDIA_TYPE_DVD_PLUS_RW:
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean
+wait_for_insertion (CDDrive *rec, gboolean blank_cd)
+{
+	CDMediaType type;
+	gboolean reload;
+
+	reload = FALSE;
+	type = guess_media_type (rec->device);
+	if (type == CD_MEDIA_TYPE_UNKNOWN) {
+		reload = TRUE;
+	}
+
+	while (!media_type_matches (type, blank_cd)) {
+		GtkWidget *dialog, *parent;
+		const char *msg, *title;
+		int res;
+
+		if (reload && blank_cd) {
+			msg = N_("Please insert a rewritable media in the drive tray.");
+			title = N_("Insert rewritable media");
+		} else if (reload && !blank_cd) {
+			msg = N_("Please insert a blank media in the drive tray.");
+			title = N_("Insert blank media");
+		} else if (blank_cd) {
+			msg = N_("Please replace the in-drive media by a rewritable media.");
+			title = N_("Reload rewritable media");
+		} else {
+			msg = N_("Please replace the in-drive media by a blank media.");
+			title = N_("Reload blank media");
+		}
+
+		parent = glade_xml_get_widget (xml, "cdr_dialog");
+		dialog = gtk_message_dialog_new (GTK_WINDOW (parent),
+				GTK_DIALOG_MODAL,
+				GTK_MESSAGE_INFO,
+				GTK_BUTTONS_OK_CANCEL,
+				"%s", _(msg));
+		gtk_dialog_set_default_response (GTK_DIALOG (dialog),
+				GTK_RESPONSE_OK);
+		gtk_window_set_title (GTK_WINDOW (dialog), _(title));
+		res = gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+		if (res == GTK_RESPONSE_CANCEL) {
+			return FALSE;
+		}
+
+		type = guess_media_type (rec->device);
+		reload = (type == CD_MEDIA_TYPE_UNKNOWN ? TRUE : FALSE);
+	}
+
+	return TRUE;
+}
+
+
 static void
 details_clicked (GtkButton *button, gpointer data)
 {
@@ -518,8 +599,12 @@ burn_cd (CDDrive *rec, const char *source_iso, const char *label, gboolean eject
 		setup_close_button ();
 		gtk_main ();
 	} else {
+		if (!wait_for_insertion (rec, blank_cd)) {
+			return RESULT_CANCEL;
+		}
+
 		cdrecorder = cd_recorder_new ();
-		
+
 		g_signal_connect (G_OBJECT (cdrecorder),
 				  "progress-changed",
 				  G_CALLBACK (progress_changed_cb),
@@ -549,8 +634,6 @@ burn_cd (CDDrive *rec, const char *source_iso, const char *label, gboolean eject
 			unlink (filename);
 		}
 
-		g_print ("res: %d\n", res);
-		
 		if (res == RESULT_FINISHED) {
 			cd_progress_set_text (_("Complete"));
 			setup_close_button ();
